import argparse
import importlib
import os
import io
import errno
import subprocess
import sys
import uuid
import json
from urllib.parse import urlparse
import tempfile
import time
from datetime import datetime, timedelta
from contextlib import redirect_stdout, redirect_stderr
import code
import random

from pyspark.sql import SparkSession, SQLContext, Row
from pyspark import SparkFiles

{% if use_instance_principle %}
USE_INSTANCE_PRINCIPLE = True
{% else %}
USE_INSTANCE_PRINCIPLE = False
OCI_CONFIG = json.loads("""{{ oci_config_str }}""")
OCI_KEY = """{{ oci_key }}"""
{% endif %}

random.seed()

def get_server_channel(region, run_dir):
    # from spark_etl.core import ServerChannelInterface
    class ServerChannel:
        def __init__(self, region, run_dir):
            # run_dir point to the current job's run dir, not the base run dir
            self.spark = None
            self.region = region
            self.run_dir = run_dir

            o = urlparse(run_dir)
            self.namespace = o.netloc.split('@')[1]
            self.bucket = o.netloc.split('@')[0]
            self.object_name_prefix = o.path[1:]  # drop the leading /


        def read_json(self, spark, name):
            from oci_core import os_download_json
            os_client = get_os_client_ex(spark, self.region)
            return os_download_json(
                os_client,
                self.namespace, self.bucket,
                os.path.join(self.object_name_prefix, name)
            )


        def has_json(self, spark, name):
            from oci_core import os_has_object
            os_client = get_os_client_ex(spark, self.region)
            return os_has_object(
                os_client,
                self.namespace, self.bucket,
                os.path.join(self.object_name_prefix, name)
            )


        def write_json(self, spark, name, payload):
            from oci_core import os_upload_json
            os_client = get_os_client_ex(spark, self.region)
            os_upload_json(
                os_client,
                payload,
                self.namespace, self.bucket,
                os.path.join(self.object_name_prefix, name)
            )


        def delete_json(self, spark, name):
            from oci_core import os_delete_object
            os_client = get_os_client_ex(spark, self.region)
            os_delete_object(
                os_client,
                self.namespace, self.bucket,
                os.path.join(self.object_name_prefix, name)
            )
    
    return ServerChannel(region, run_dir)

# lib installer
def _install_libs(run_id):
    current_dir = os.getcwd()
    base_dir    = os.path.join(current_dir, run_id)
    lib_dir     = os.path.join(base_dir, 'python_libs')
    lib_zip     = SparkFiles.get("lib.zip")
    lock_name   = os.path.join(base_dir, '__lock__')

    os.makedirs(base_dir, exist_ok=True)

    for i in range(0, 100):
        try:
            lock_fh = os.open(lock_name, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
            os.close(lock_fh)
            try:
                if not os.path.isdir(lib_dir):
                    print("_install_libs: install lib starts")
                    os.makedirs(lib_dir)
                    subprocess.check_call(['unzip', "-qq", lib_zip, "-d", lib_dir])
                    print("_install_libs: install lib done")
                if lib_dir not in sys.path:
                    print(f"_install_libs: add {lib_dir} path")
                    sys.path.insert(0, lib_dir)
                return
            finally:
                os.remove(lock_name)
        except OSError as e:
            if e.errno == errno.EEXIST:
                time.sleep(random.randint(1, 10))
                continue
            raise

    raise Exception("Failed to install libraries!")

class Asker:
    def __init__(self, run_dir, app_region):
        self.spark = None
        self.run_dir = run_dir
        self.app_region = app_region

    def initialize(self, spark):
        from oci_core import dfapp_get_os_client, get_delegation_token, get_os_client

        if USE_INSTANCE_PRINCIPLE:
            delegation_token = get_delegation_token(spark)
            self.os_client = dfapp_get_os_client(self.app_region, delegation_token)
        else:
            with tempfile.NamedTemporaryFile(mode='w+t', delete=False) as key_f:
                key_f.write(OCI_KEY)
            _oci_config = dict(OCI_CONFIG)
            _oci_config['key_file'] = key_f.name
            self.os_client = get_os_client(None, config=_oci_config)

    # place a request to the launcher
    def __call__(self, content, timeout=timedelta(minutes=10)):
        from oci_core import os_upload_json, os_download_json

        request_id = str(uuid.uuid4())

        o = urlparse(self.run_dir)
        namespace = o.netloc.split('@')[1]
        bucket = o.netloc.split('@')[0]
        object_name_prefix = o.path[1:]  # drop the leading /

        os_upload_json(self.os_client, content, namespace, bucket, f"{object_name_prefix}/to_submitter/ask_{request_id}.json")
        answer_name = f"{object_name_prefix}/to_submitter/answer_{request_id}.json"
        start_time = datetime.utcnow()
        while True:
            r = self.os_client.list_objects(
                namespace,
                bucket,
                prefix = answer_name,
                limit = 1
            )
            if len(r.data.objects) == 0:
                if (datetime.utcnow() - start_time) >= timeout:
                    raise Exception("Ask is not answered: timed out")
                time.sleep(5)
                continue

            return os_download_json(self.os_client, namespace, bucket, answer_name)


def get_os_client_ex(spark, region):
    from oci_core import dfapp_get_os_client, get_delegation_token, get_os_client
    if USE_INSTANCE_PRINCIPLE:
        delegation_token = get_delegation_token(spark)
        os_client = dfapp_get_os_client(region, delegation_token)
    else:
        with tempfile.NamedTemporaryFile(mode='w+t', delete=False) as key_f:
            key_f.write(OCI_KEY)
        _oci_config = dict(OCI_CONFIG)
        _oci_config['key_file'] = key_f.name
        os_client = get_os_client(None, config=_oci_config)
    return os_client


def _bootstrap():
    parser = argparse.ArgumentParser(description='job')
    parser.add_argument(
        "--deployment-location", type=str, required=True, help="Deployment Location",
    )
    parser.add_argument(
        "--run-id", type=str, required=True, help="Run ID",
    )
    parser.add_argument(
        "--run-dir", type=str, required=True, help="Run Directory",
    )
    parser.add_argument(
        "--app-region", type=str, required=True, help="Application Region",
    )
    args = parser.parse_args()

    o = urlparse(args.run_dir)
    if o.scheme != 'oci':
        raise Exception("run_dir MUST be in OCI")

    # Load application
    spark = SparkSession.builder.appName("RunJob").getOrCreate()
    sc = spark.sparkContext
    sc.addPyFile(f"{args.deployment_location}/app.zip")
    sc.addFile(f"{args.deployment_location}/lib.zip")
    print("Application loaded")

    # The archive file goes into /opt/dataflow
    # Load python libraries
    current_dir = os.getcwd()
    lib_dir = os.path.join(current_dir, 'python_libs')
    os.mkdir(lib_dir)
    lib_zip = SparkFiles.get("lib.zip")
    subprocess.call([
        'unzip', "-qq", lib_zip, "-d", lib_dir
    ])
    sys.path.insert(0, lib_dir)

    run_id = args.run_id
    run_dir = args.run_dir

    server_channel = get_server_channel(args.app_region, run_dir)
    xargs = server_channel.read_json(spark, "args.json")

    entry = importlib.import_module("main")
    result = entry.main(spark, xargs, sysops={
        "install_libs": lambda : _install_libs(run_id),
        "ask": Asker(run_dir, args.app_region),
        "channel": server_channel
    })
    server_channel.write_json(spark, "result.json", result)


class PySparkConsole(code.InteractiveInterpreter):
    def __init__(self, locals=None):
        super(PySparkConsole, self).__init__(locals=locals)



_bootstrap()



import json
import os
import uuid
import tempfile
import subprocess
import json
import time

from spark_etl.job_submitters import AbstractJobSubmitter
from spark_etl.core import ClientChannelInterface
from spark_etl.utils import CLIHandler

class ClientChannel(ClientChannelInterface):
    def __init__(self, run_dir, run_id):
        self.run_dir = run_dir
        self.run_id = run_id

    def _get_json_path(self, name):
        return os.path.join(self.run_dir, self.run_id, name)

    def read_json(self, name):
        with open(self._get_json_path(name), "r") as f:
            return json.load(f)

    def has_json(self, name):
        return os.path.isfile(self._get_json_path(name))

    def write_json(self, name, payload):
        with open(self._get_json_path(name), "w") as f:
            json.dump(payload, f)

    def delete_json(self, name):
        os.remove(self._get_json_path(name))


class PySparkJobSubmitter(AbstractJobSubmitter):
    def __init__(self, config):
        super(PySparkJobSubmitter, self).__init__(config)
        ####################################################################################
        # here is a sample config
        # {
        #     "run_dir": "/home/stonezhong/spark-etl-lab/src/local-lake/runs",
        #     "enable_aws_s3": true,
        #     "aws_account": "~/.aws/account.json",
        #     "aws_s3_buffer_dir": "/home/stonezhong/spark-etl-lab/src/local-lake/s3-buffer"
        # }
        # enable_aws_s3: set to True if you want to access AWS S3 buckets
        # aws_account: point to a json filename, the json file looks like below
        # {
        #     "aws_access_key_id": "***",
        #     "aws_secret_access_key": "***"
        # }
        # aws_s3_buffer_dir: a temp dir for s3 buffered data
        ####################################################################################



    def run(self, deployment_location, options={}, args={}, handlers=[], on_job_submitted=None, cli_mode=False):
        # version is already baked into deployment_location
        # local submitter ignores handlers
        run_id  = str(uuid.uuid4())
        run_dir = self.config['run_dir']
        app_dir = deployment_location

        os.makedirs(os.path.join(run_dir, run_id))

        # generate input.json
        with open(os.path.join(run_dir, run_id, 'input.json'), 'wt') as f:
            json.dump(args, f)

        client_channel = ClientChannel(run_dir, run_id)

        run_args = [
            "spark-submit",
        ]
        enable_aws_s3 = self.config.get('enable_aws_s3', False)
        if enable_aws_s3:
            # spark will automatically download package
            run_args.extend([
                "--packages",
                "org.apache.hadoop:hadoop-aws:2.7.3"
            ])

        run_args.extend([
            os.path.join(deployment_location, "job_loader.py"),
            "--run-id", run_id,
            "--run-dir", run_dir,
            "--app-dir", app_dir
        ])

        if enable_aws_s3:
            run_args.append("--enable-aws-s3")
            aws_s3_buffer_dir = self.config.get('aws_s3_buffer_dir')
            if aws_s3_buffer_dir is not None:
                aws_s3_buffer_dir = os.path.expandvars(os.path.expanduser(aws_s3_buffer_dir))
                run_args.extend(['--aws-s3-buffer-dir', aws_s3_buffer_dir])


        aws_account = self.config.get('aws_account')
        if aws_account is not None:
            aws_account = os.path.expandvars(os.path.expanduser(aws_account))
            run_args.extend(['--aws-account', aws_account])

        p = subprocess.Popen(run_args)
        exit_code = None
        cli_entered = False

        if on_job_submitted is not None:
            on_job_submitted(run_id, vendor_info={})

        while True:
            time.sleep(1)
            exit_code = p.poll()
            if exit_code is not None:
                break

            if cli_mode and not cli_entered:
                cli_entered = True
                cli_handler = CLIHandler(
                    client_channel,
                    lambda : p.poll() is None,
                    handlers
                )
                cli_handler.loop()

        if exit_code != 0:
            Exception(f"Job failed, exit_code = {exit_code}")

        print("Job completed successfully")
        with open(os.path.join(run_dir, run_id, "result.json"), "r") as f:
            return json.load(f)



import importlib
import argparse
import argparse
import uuid
import os
import subprocess
import sys
import tempfile
import json
import random

from pyspark import SparkFiles
from pyspark.sql import SparkSession

random.seed()

def get_server_channel(run_dir):
    # from spark_etl.core import ServerChannelInterface
    class ServerChannel:
        def __init__(self, run_dir):
            self.run_dir = run_dir

        def read_json(self, spark, name):
            stage_file = tempfile.NamedTemporaryFile(mode="w", delete=False)
            stage_file.close()
            try:
                subprocess.check_call([
                    "hdfs", "dfs", "-copyToLocal", "-f",
                    os.path.join(self.run_dir, name),
                    stage_file.name
                ])
                with open(stage_file.name) as f:
                    return json.load(f)
            finally:
                os.remove(stage_file.name)


        def has_json(self, spark, name):
            exit_code = subprocess.call([
                "hdfs", "dfs", "-test", "-f",
                os.path.join(self.run_dir, name)
            ])
            if exit_code == 0:
                return True
            if exit_code == 1:
                return False
            raise Exception(f"Unrecognized exit_code: {exit_code}")


        def write_json(self, spark, name, payload):
            stage_file = tempfile.NamedTemporaryFile(mode="w", delete=False)
            stage_file.close()
            try:
                with open(stage_file.name, "wt") as f:
                    json.dump(payload, f)
                subprocess.check_call([
                    "hdfs", "dfs", "-copyFromLocal", stage_file.name,
                    os.path.join(self.run_dir, name)
                ])
            finally:
                os.remove(stage_file.name)


        def delete_json(self, spark, name):
            subprocess.check_call([
                "hdfs", "dfs", "-rm", os.path.join(self.run_dir, name)
            ])

    return ServerChannel(run_dir)


# lib installer
def _install_libs(run_id):
    ##########################################
    #
    # |
    # +-- {lib_dir}
    #       |
    #       +-- uuid1.zip
    #       |
    #       +-- uuid2.zip
    #       |
    #       +-- uuid1  (lib extracted)
    #       |
    #       +-- uuid2  (lib extracted)
    ##########################################
    print("job_loader._install_libs: enter, run_id = {}".format(run_id))

    current_dir = os.path.dirname(os.path.abspath(__file__))
    print(f"current_dir = {current_dir}")

    base_dir    = os.path.join(current_dir, run_id)
    lib_dir     = os.path.join(base_dir, 'python_libs')
    lib_zip     = SparkFiles.get("lib.zip")
    lock_name   = os.path.join(base_dir, '__lock__')

    os.makedirs(base_dir, exist_ok=True)

    for i in range(0, 100):
        try:
            lock_fh = os.open(lock_name, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
            os.close(lock_fh)
            try:
                if not os.path.isdir(lib_dir):
                    print("_install_libs: install lib starts")
                    os.makedirs(lib_dir)
                    subprocess.check_call(['unzip', "-qq", lib_zip, "-d", lib_dir])
                    print("_install_libs: install lib done")
                if lib_dir not in sys.path:
                    print(f"_install_libs: add {lib_dir} path")
                    sys.path.insert(0, lib_dir)
                print("job_loader._install_libs: exit")
                return
            finally:
                os.remove(lock_name)
        except OSError as e:
            if e.errno == errno.EEXIST:
                time.sleep(random.randint(1, 10))
                continue
            raise

    raise Exception("Failed to install libraries!")



print("job_loader.py: enter")

parser = argparse.ArgumentParser(description='job')
parser.add_argument(
    "--run-id", type=str, required=True, help="Run ID",
)
parser.add_argument(
    "--run-dir", type=str, required=True, help="Run Directory",
)
parser.add_argument(
    "--lib-zip", type=str, required=True, help="Zipped library",
)

args = parser.parse_args()
spark = SparkSession.builder.appName("RunJob").getOrCreate()

sc = spark.sparkContext
sc.addFile(args.lib_zip)

_install_libs(args.run_id)

# get input
server_channel = get_server_channel(args.run_dir)
input_args = server_channel.read_json(spark, "input.json")

try:
    entry = importlib.import_module("main")
    result = entry.main(spark, input_args, sysops={
        "install_libs": lambda : _install_libs(run_id),
        "channel": get_server_channel(args.run_dir)
    })

    # save output
    server_channel.write_json(spark, "result.json", result)
    print("job_loader: exit gracefully")
finally:
    spark.stop()



import json

class Build:
    """ This represent an application build """
    def __init__(self, build_dir):
        """
        Parameters
        ----------
        build_dir: str
            The directory for the build.
        """
        self.build_dir = build_dir
        with open(f"{self.build_dir}/manifest.json", "r") as f:
            self.manifest = json.load(f)

    @property
    def artifacts(self):
        """List of artifacts of the build"""
        return ("app.zip", "lib.zip", "main.py", "manifest.json")

    @property
    def version(self):
        """Build version"""
        return self.manifest["version"]



import os
import tempfile
import shutil
import subprocess
import socket
from contextlib import closing
import random
import time

random.seed()

_DEBUG_SSH = False

class SSHConfig:
    def __init__(self, config):
        self.config = config
        self.home_dir = None

    def line_to_key_value(self, line):
        pos = line.find(' ')
        key = line[:pos]
        value = line[pos+1:]
        return key, value

    def transform_line_host(self, value):
        return [
            f"Host {value}",
            "    StrictHostKeyChecking no",
            "    UserKnownHostsFile /dev/null"
        ]

    def transform_line_identity_file(self, value):
        return [
            f"    IdentityFile {self.home_dir}/keys/{value}"
        ]

    def transform_line_proxy_command(self, value):
        if not value.startswith("ssh "):
            return [
                f"    ProxyCommand {value}"
            ]
        return [
            f"    ProxyCommand ssh -F {os.path.join(self.home_dir, 'config')} {value[4:]}"
        ]

    def transform_line(self, line):
        key, value = self.line_to_key_value(line)
        ret = []

        if key == 'Host':
            return self.transform_line_host(value)

        if key == "IdentityFile":
            return self.transform_line_identity_file(value)

        if key == 'ProxyCommand':
            return self.transform_line_proxy_command(value)

        return [f"    {key} {value}"]


    def generate(self):
        self.home_dir = tempfile.mkdtemp(suffix="-ssh")
        os.mkdir(os.path.join(self.home_dir, "keys"))

        keys = self.config["keys"]
        config = self.config['config']

        # save all keys
        for k, v in keys.items():
            key_filename = os.path.join(self.home_dir, "keys", k)
            with open(key_filename, "wt") as kf:
                kf.write(v)
            os.chmod(key_filename, 0o600)

        config_filename = os.path.join(self.home_dir, "config")
        with open(config_filename, "wt") as ssh_cfg_f:
            effective_lines = []
            for line in config.split("\n"):
                l = line.strip()
                if len(l) == 0 or l[0] == '#':
                    continue
                effective_lines.append(l)

            for line in effective_lines:
                ls = self.transform_line(line)
                for l in ls:
                    print(l, file=ssh_cfg_f)
        os.chmod(config_filename, 0o664)


    def destroy(self):
        if self.home_dir is not None:
            shutil.rmtree(self.home_dir)

    def get_scp_cmds(self):
        return [
            "scp",
            "-F", f"{os.path.join(self.home_dir, 'config')}",
            "-q",
        ]

    def get_ssh_cmds(self):
        return [
            "ssh",
            "-F", f"{os.path.join(self.home_dir, 'config')}",
            "-q",
        ]

    def execute(self, host, cmds, error_ok=False, retry_count=5, wait_interval=5):
        for i in range(0, retry_count):
            ret = self._execute(host, cmds)
            if ret == 255:
                # this means the command is not even executed on remote host
                time.sleep(random.randrange(wait_interval)+1)
                continue
            if ret == 0 or error_ok:
                return ret
            raise Exception(f"{' '.join(cmds)} failed on host {host} with exit code {ret}")
        raise Exception(f"{' '.join(cmds)} failed on host {host} with exit code {ret}")

    def _execute(self, host, cmds):
        ssh_cmds = self.get_ssh_cmds()
        ssh_cmds.append(host)
        ssh_cmds.extend(cmds)
        if _DEBUG_SSH:
            print(f"running {' '.join(ssh_cmds)}")
        ret = subprocess.call(
            ssh_cmds,
            stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL,
            shell=False
        )
        return ret


    def scp(self, src, target, error_ok=False):
        scp_cmds = self.get_scp_cmds()
        scp_cmds.extend([src, target])
        if _DEBUG_SSH:
            print(f"running {' '.join(scp_cmds)}")
        if error_ok:
            subprocess.call(
                scp_cmds,
                stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL,
                shell=False
            )
        else:
            subprocess.check_call(
                scp_cmds,
                stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL,
                shell=False
            )

    def find_free_local_tcp_port(self):
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(('127.0.0.1', 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            return s.getsockname()[1]

    # create a TCP port forwarding
    def tunnel(self, host, remote_host, remote_port):
        local_port = self.find_free_local_tcp_port()
        ssh_cmds = self.get_ssh_cmds()
        ssh_cmds.extend([
            "-N",
            "-L",
            f"{local_port}:{remote_host}:{remote_port}",
            host
        ])
        if _DEBUG_SSH:
            print(f"running {' '.join(ssh_cmds)}")
        proc = subprocess.Popen(
            ssh_cmds,
            stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL,
            shell=False
        )
        return local_port, proc




from .abstract_deployer import AbstractDeployer

# User should cherry pick deployers from the deployer files
# for example:
# from spark_etl.deployers.s3_deployer import S3Deployer
#
# Since different deployer probably depened on different
# vendor library, putting all those deployers here means if user import
# spark_etl.deployers, they will import all vendor's library (such as boto3, oci, etc)
# which is not necessary. Most of time, user only need 1 vendor.



import os
import json
import time
import subprocess
from urllib.parse import urlparse
import uuid
import tempfile

from requests.auth import HTTPBasicAuth
import requests

from .abstract_job_submitter import AbstractJobSubmitter
from spark_etl.utils import CLIHandler
from spark_etl.core import ClientChannelInterface
from spark_etl.exceptions import SparkETLLaunchFailure
from spark_etl.ssh_config import SSHConfig

#################################################################################
# For ssh, we allow user to specify their own configs
# The config file is normal ssh config file, however, the IdentityFile should
# point to a key file id instead of filename
#
# The entire config is a json, which:
# {
#     config: string, same as ssh config,
#     keys: {
#         "foo": "...",         # content for key "foo"
#         "bar": "..."          # content for key "bar"
#     }
# }
#################################################################################



class ClientChannel(ClientChannelInterface):
    def __init__(self, bridge, stage_dir, run_dir, run_id, ssh_config):
        self.bridge = bridge        # the bridge server's name which we can ssh to
        self.stage_dir = stage_dir  # stage_dir is on bridge
        self.run_dir = run_dir      # base dir for all runs, e.g. hdfs:///beta/etl/runs
        self.run_id  = run_id       # string, unique id of the run
        self.ssh_config = ssh_config


    def _create_stage_dir(self):
        # We are using stage_dir instead of temp dir to stage objects on bridge
        # the reason is, many cases temp dir is in memory and cannot hold large object
        self.ssh_config.execute(self.bridge, [
            "mkdir", "-p", os.path.join(self.stage_dir, self.run_id)
        ])


    def read_json(self, name):
        """read a json object from HDFS run dir with specific name
        """
        self._create_stage_dir()
        self.ssh_config.execute(self.bridge, [
            "hdfs", "dfs", "-copyToLocal", "-f",
            os.path.join(self.run_dir, self.run_id, name),
            os.path.join(self.stage_dir, self.run_id, name)
        ])

        try:
            stage_file = tempfile.NamedTemporaryFile(mode="w", delete=False)
            stage_file.close()
            self.ssh_config.scp(
                os.path.join(
                    f"{self.bridge}:{self.stage_dir}", self.run_id, name
                ),
                stage_file.name
            )
            with open(stage_file.name) as f:
                return json.load(f)
        finally:
            os.remove(stage_file.name)


    def has_json(self, name):
        """check if a file with specific name exist or not in the HDFS run dir
        """
        exit_code = self.ssh_config.execute(self.bridge, [
            "hdfs", "dfs", "-test", "-f",
            os.path.join(self.run_dir, self.run_id, name)
        ], error_ok=True)
        if exit_code == 0:
            return True
        if exit_code == 1:
            return False
        raise Exception(f"Unrecognized exit code: {exit_code}")


    def write_json(self, name, payload):
        """write a json object to HDFS run dir with specific name
        """
        self._create_stage_dir()

        stage_file = tempfile.NamedTemporaryFile(mode="w", delete=False)
        stage_file.close()
        try:
            with open(stage_file.name, "wt") as f:
                json.dump(payload, f)

            # copy over to bridge at stage directory
            self.ssh_config.scp(
                stage_file.name,
                os.path.join(
                    f"{self.bridge}:{self.stage_dir}", self.run_id, name
                )
            )

            # then upload to HDFS, -f for overwrite if exist
            self.ssh_config.execute(self.bridge, [
                "hdfs", "dfs", "-copyFromLocal", "-f",
                os.path.join(self.stage_dir, self.run_id, name),
                os.path.join(self.run_dir, self.run_id, name)
            ])
        finally:
            os.remove(stage_file.name)


    def delete_json(self, name):
        self.ssh_config.execute(self.bridge, [
            "hdfs", "dfs", "-rm",
            os.path.join(self.run_dir, self.run_id, name)
        ])


class LivyJobSubmitter(AbstractJobSubmitter):
    def __init__(self, config):
        super(LivyJobSubmitter, self).__init__(config)
        self.ssh_config = SSHConfig(config['ssh_config'])

    # options is vendor specific arguments
    # args.conf is the spark job config
    # args is arguments need to pass to the main entry
    def run(self, deployment_location, options={}, args={}, handlers=[], on_job_submitted=None, cli_mode=False):
        bridge = self.config["bridge"]
        self.ssh_config.generate()
        tunnel = None
        local_livy_port = None

        # by default, livy server is the same as the bridge
        # and ssh tunnel is used for livy access
        livy_cfg = self.config.get('livy', {
            'host': '127.0.0.1',
            'port': 8998,
            'via_tunnel': True,
            'protocol': 'http',
            'username': None,
            'password': None,
        })
        livy_cfg_host = livy_cfg.get('host', '127.0.0.1')
        livy_cfg_port = livy_cfg.get('port', 8998)
        livy_cfg_via_tunnel = livy_cfg.get('via_tunnel', True)
        livy_cfg_protocol = livy_cfg.get('protocol', 'http')
        livy_cfg_username = livy_cfg.get('username', None)
        livy_cfg_password = livy_cfg.get('password', None)
        try:
            if livy_cfg_via_tunnel:
                local_livy_port, tunnel = self.ssh_config.tunnel(bridge, livy_cfg_host, livy_cfg_port)
            o = urlparse(deployment_location)
            if o.scheme not in ('hdfs', 's3', 's3a'):
                raise SparkETLLaunchFailure("deployment_location must be in hdfs, s3 or s3a")


            stage_dir = self.config['stage_dir']
            run_dir = self.config['run_dir']

            # create run dir
            run_id = str(uuid.uuid4())
            self.ssh_config.execute(bridge, [
                "hdfs", "dfs", "-mkdir", "-p",
                os.path.join(run_dir, run_id)
            ])

            client_channel = ClientChannel(
                bridge, stage_dir, run_dir, run_id,
                ssh_config = self.ssh_config
            )
            client_channel.write_json("input.json", args)


            headers = {
                "Content-Type": "application/json",
                "X-Requested-By": "root",
                'proxyUser': 'root'
            }

            config = {
                'file': os.path.join(deployment_location, "job_loader.py"),
                'pyFiles': [ os.path.join(deployment_location, "app.zip") ],
                'args': [
                    '--run-id', run_id,
                    '--run-dir', os.path.join(run_dir, run_id),
                    '--lib-zip', os.path.join(deployment_location, "lib.zip")
                ]
            }

            config.update(options)
            config.pop("display_name", None)  # livy job submitter does not support display_name

            if livy_cfg_via_tunnel:
                livy_service_url = f"{livy_cfg_protocol}://127.0.0.1:{local_livy_port}/"
            else:
                livy_service_url = f"{livy_cfg_protocol}://{livy_cfg_host}:{livy_cfg_port}/"

            print(f"Livy: {livy_service_url}")
            # print(json.dumps(config))
            if livy_cfg_username is None:
                r = requests.post(
                    os.path.join(livy_service_url, "batches"),
                    data=json.dumps(config),
                    headers=headers,
                    verify=False
                )
            else:
                r = requests.post(
                    os.path.join(livy_service_url, "batches"),
                    data=json.dumps(config),
                    headers=headers,
                    auth=HTTPBasicAuth(livy_cfg_username, livy_cfg_password),
                    verify=False
                )
            if r.status_code not in [200, 201]:
                msg = "Failed to submit the job, status: {}, error message: \"{}\"".format(
                    r.status_code,
                    r.content
                )
                raise SparkETLLaunchFailure(msg)

            print(f'job submitted, run_id = {run_id}')
            ret = json.loads(r.content.decode("utf8"))
            if on_job_submitted is not None:
                on_job_submitted(run_id, vendor_info=ret)
            job_id = ret['id']
            print('job id: {}'.format(job_id))
            print(ret)
            print('logs:')
            for log in ret.get('log', []):
                print(log)

            def get_job_status():
                if livy_cfg_username is None:
                    r = requests.get(
                        os.path.join(livy_service_url, "batches", str(job_id)),
                        headers=headers,
                        verify=False
                    )
                else:
                    r = requests.get(
                        os.path.join(livy_service_url, "batches", str(job_id)),
                        headers=headers,
                        auth=HTTPBasicAuth(livy_cfg_username, livy_cfg_password),
                        verify=False
                    )
                if r.status_code != 200:
                    msg = "Failed to get job status, status: {}, error message: \"{}\"".format(
                        r.status_code,
                        r.content
                    )
                    # this is considered unhandled
                    raise Exception(msg)
                ret = json.loads(r.content.decode("utf8"))
                return ret

            cli_entered = False
            # pull the job status
            while True:
                ret = get_job_status()
                job_state = ret['state']
                appId = ret.get('appId')
                print('job_state: {}, applicationId: {}'.format(
                    job_state, appId

                ))
                if job_state in ['success', 'dead', 'killed']:
                    # cmd = f"yarn logs -applicationId {appId}"
                    # host = self.config['bridge']
                    # subprocess.call(["ssh", "-q", "-t", host, cmd], shell=False)
                    print(f"job_state={job_state}")
                    if job_state == 'success':
                        return client_channel.read_json("result.json")
                    raise SparkETLLaunchFailure("Job failed")
                time.sleep(5)

                # we enter the cli mode upon first reached the running status
                if cli_mode and not cli_entered and job_state == 'running':
                    cli_entered = True
                    cli_handler = CLIHandler(
                        client_channel,
                        lambda : get_job_status()['state'] not in ('success', 'dead', 'killed'),
                        handlers
                    )
                    cli_handler.loop()
        finally:
            if tunnel is not None:
                tunnel.kill()
            self.ssh_config.destroy()





class SparkETLException(Exception):
    """Base class for exception raised from spark_etl package"""
    def __init__(self, message, cause=None):
        super(SparkETLException, self).__init__(message)
        self.cause = cause

class SparkETLDeploymentFailure(SparkETLException):
    """spark-etl failed to deploy an application"""
    def __init__(self, message, cause=None):
        super(SparkETLDeploymentFailure, self).__init__(message, cause=cause)

class SparkETLLaunchFailure(SparkETLException):
    """spark-etl failed to launch an application"""
    def __init__(self, message, cause=None):
        super(SparkETLLaunchFailure, self).__init__(message, cause=cause)




import uuid
import subprocess
import os
from urllib.parse import urlparse

from .abstract_deployer import AbstractDeployer
from spark_etl import Build
from spark_etl.exceptions import SparkETLDeploymentFailure
from spark_etl.ssh_config import SSHConfig


class HDFSDeployer(AbstractDeployer):
    """
    This deployer deploys application to HDFS
    """
    def __init__(self, config):
        super(HDFSDeployer, self).__init__(config)
        self.ssh_config = SSHConfig(config['ssh_config'])

    def deploy(self, build_dir, deployment_location):
        self.ssh_config.generate()
        try:
            o = urlparse(deployment_location)
            if o.scheme not in ('hdfs', 's3', 's3a'):
                raise SparkETLDeploymentFailure("deployment_location must be in hdfs, s3 or s3a")

            # let's copy files to the stage dir
            bridge_dir = os.path.join(self.config['stage_dir'], str(uuid.uuid4()))

            bridge = self.config["bridge"]
            self.ssh_config.execute(bridge, ["mkdir", "-p", bridge_dir])

            build = Build(build_dir)

            for artifact in build.artifacts:
                self.ssh_config.scp(
                    f"{build_dir}/{artifact}", f"{bridge}:{bridge_dir}/{artifact}"
                )

            # copy job loader
            job_loader_filename = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'job_loader.py'
            )
            self.ssh_config.scp(job_loader_filename, f"{bridge}:{bridge_dir}/job_loader.py")

            dest_location = f"{deployment_location}/{build.version}"
            self.ssh_config.execute(bridge, ["hdfs", "dfs", "-rm", "-r", dest_location], error_ok=True)
            self.ssh_config.execute(bridge, ["hdfs", "dfs", "-mkdir", "-p", dest_location])

            artifacts = []
            artifacts.extend(build.artifacts)
            artifacts.append("job_loader.py")
            for artifact in artifacts:
                self.ssh_config.execute(bridge, [
                    "hdfs", "dfs", "-copyFromLocal", f"{bridge_dir}/{artifact}", f"{dest_location}/{artifact}"
                ])

            self.ssh_config.execute(bridge, ["rm", "-rf", bridge_dir])
        finally:
            self.ssh_config.destroy()



import uuid
import os
import subprocess
from urllib.parse import urlparse
import json
import tempfile

from jinja2 import Template

import oci
from oci_core import get_os_client, get_df_client, os_upload, os_upload_json

from spark_etl.deployers import AbstractDeployer
from spark_etl import Build, SparkETLDeploymentFailure
from .tools import check_response, remote_execute

def _save_json_temp(payload):
    # save a dict to temporary json file, and return filename
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(json.dumps(payload).encode('utf8'))
        return f.name

def get_job_loader(oci_config):
    # render job_loader.py, rendered result in a temporary file
    # return the rendered filename
    job_loader_filename = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'job_loader.py'
    )

    if oci_config is not None:
        oci_cfg = dict(oci_config)
        key_file = oci_cfg.pop("key_file")
        with open(key_file, "rt") as key_f:
            oci_key = key_f.read()

    with open(job_loader_filename, "rt") as f:
        load_content = f.read()
        template = Template(load_content)
        if oci_config is None:
            c = template.render(
                use_instance_principle = True,
            )
        else:
            c = template.render(
                use_instance_principle = False,
                oci_config_str = json.dumps(oci_cfg, indent=4),
                oci_key = oci_key
            )
        with tempfile.NamedTemporaryFile(mode='w+t', delete=False) as f:
            f.write(c)
            return f.name


class DataflowDeployer(AbstractDeployer):
    def __init__(self, config):
        super(DataflowDeployer, self).__init__(config)
        # config must have
        #   region, e.g., value is IAD
        #   compartment_id: the compartment we want to have the dataflow app
        #   driver_shape
        #   executor_shape
        #   num_executors
        #   spark_version

    @property
    def region(self):
        return self.config["region"]

    def create_application(self, manifest, destination_location):
        df_client = get_df_client(self.region, config=self.config.get("oci_config"))
        dataflow = self.config['dataflow']

        display_name = f"{manifest['display_name']}-{manifest['version']}"

        # if the application already exist, we will fail the deployment
        # since user should bump the version for new deployment
        r = df_client.list_applications(
            dataflow['compartment_id'],
            limit=1,
            display_name=display_name,
        )
        check_response(r, lambda : SparkETLDeploymentFailure("Unable to list application"))
        if len(r.data) > 0:
            raise SparkETLDeploymentFailure(f"Application {display_name} already created")

        create_application_details = oci.data_flow.models.CreateApplicationDetails(
            compartment_id=dataflow['compartment_id'],
            display_name=display_name,
            driver_shape=dataflow['driver_shape'],
            executor_shape=dataflow['executor_shape'],
            num_executors=dataflow['num_executors'],
            spark_version=dataflow.get("spark_version", "2.4.4"),
            file_uri=f"{destination_location}/{manifest['version']}/job_loader.py",
            language="PYTHON",
            archive_uri=dataflow.get("archive_uri")
        )

        r = df_client.create_application(
            create_application_details=create_application_details
        )
        check_response(r, lambda : SparkETLDeploymentFailure("Unable to create dataflow application"))
        return r.data

    def deploy(self, build_dir, destination_location):
        o = urlparse(destination_location)
        if o.scheme != 'oci':
            raise SparkETLDeploymentFailure("destination_location must be in OCI")

        namespace = o.netloc.split('@')[1]
        bucket = o.netloc.split('@')[0]
        root_path = o.path[1:]    # remove the leading "/"

        build = Build(build_dir)

        print("Uploading files:")
        # Data flow want to call python lib python.zip
        os_client = get_os_client(self.region, config=self.config.get("oci_config"))
        for artifact in build.artifacts:
            os_upload(
                os_client,
                f"{build_dir}/{artifact}",
                namespace,
                bucket,
                f"{root_path}/{build.version}/{artifact}"
            )

        # let's upload the job loader
        job_loader_filename = get_job_loader(self.config.get("oci_config"))

        os_upload(
            os_client,
            job_loader_filename,
            namespace,
            bucket,
            f"{root_path}/{build.version}/job_loader.py"
        )

        application = self.create_application(build.manifest, destination_location)
        app_info = {
            "application_id": application.id,
            "compartment_id": application.compartment_id
        }

        os_upload_json(
            os_client, app_info,
            namespace, bucket, f"{root_path}/{build.version}/deployment.json"
        )

        oci_config = self.config.get("oci_config")
        if oci_config is not None:
            os_upload(
                os_client,
                _save_json_temp(oci_config),
                namespace,
                bucket,
                "oci_config.json"
            )
            os_upload(
                os_client,
                oci_config['key_file'],
                namespace,
                bucket,
                "oci_api_key.pem",
            )





import io
import subprocess
import time
from contextlib import redirect_stdout, redirect_stderr
import code
from datetime import datetime
import readline
import traceback

from termcolor import colored, cprint

CLI_REQUEST_NAME  = "cli-request.json"
CLI_RESPONSE_NAME = "cli-response.json"
HELP_TEXT         = """\
******************************************************
*                                                    *
*              Spark-CLI command list                *
*                                                    *
******************************************************

@@help             -- print help
@@bash             -- execute the buffer as BASH batch
@@python           -- execute the buffer as python
@@log <filename>   -- log to local file
@@clear            -- clear buffer
@@show             -- show buffer
@@load <filename>  -- append content of file to the buffer
@@mode off         -- turn off line mode
@@mode bash        -- turn on line mode, each line is bash command
@@mode python      -- turn on line mode, each line is a python code
@@quit             -- quit Spark-CLI
"""

def handle_bash(spark, user_input, channel):
    cmd_buffer = '\n'.join(user_input['lines'])
    f = io.StringIO()
    with redirect_stdout(f):
        p = subprocess.run(cmd_buffer, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    channel.write_json(
        spark,
        CLI_RESPONSE_NAME,
        {
            "status": "ok",
            "exit_code": p.returncode,
            "output": p.stdout.decode('utf-8'),
        }
    )

def handle_python(spark, user_input, console, channel):
    source = '\n'.join(user_input['lines'])
    stdout_f = io.StringIO()
    stderr_f = io.StringIO()
    with redirect_stdout(stdout_f):
        with redirect_stderr(stderr_f):
            console.runsource(source, symbol="exec")

    channel.write_json(
        spark,
        CLI_RESPONSE_NAME,
        {
            "status": "ok",
            "output": stdout_f.getvalue() + "\n" + stderr_f.getvalue() ,
        }
    )

class PySparkConsole(code.InteractiveInterpreter):
    def __init__(self, locals=None):
        super(PySparkConsole, self).__init__(locals=locals)


class CLIHandler:
    def __init__(self, client_channel, is_job_active, handlers):
        self.client_channel = client_channel
        if is_job_active is None:
            self.is_job_active = lambda : True
        else:
            self.is_job_active = is_job_active

        self.handlers = handlers
        self.last_job_ok_time = None


    def loop(self):
        # line_mode can be "bash", "python" or "OFF"
        # When line_mode is OFF, you need send explicitly run @@bash or @@python
        # to submit a block of code to server
        # When line_mode is bash, each line is a bash script
        # when line_mode is python, each line is a python script
        line_mode = "off"

        # if is_waiting_for_response is True, we need to pull server for cli-response.json
        # if is_waiting_for_response is False, we are free to enter new command
        is_waiting_for_response = True
        # command line buffer
        cli_lines = []
        cli_wait_prompt = "-/|\\"
        cli_wait_prompt_idx = 0
        log_filename = None
        # commands
        # @@log           -- all the output will be written to the log file as well
        #                    by default there is not log
        # @@nolog         -- turn off log
        # @@clear         -- clear command line buffer
        # @@load          -- load a script from local file and append to command buffer
        # @@bash          -- submit a bash script
        # @@python        -- submit a python script
        # @@show          -- show the command buffer
        # @@pwd           -- show driver's current directory
        # @@quit          -- quit the cli console

        command = None
        while True:
            handle_server_ask(self.client_channel, self.handlers)
            if not is_waiting_for_response:
                if line_mode == "off":
                    prompt = "> "
                elif line_mode == "bash":
                    prompt = "bash> "
                else:
                    prompt = "python> "

                command = input(prompt)

                if command == "@@help":
                    print(HELP_TEXT)
                    print("")
                    continue

                if command == "@@quit":
                    self.client_channel.write_json(CLI_REQUEST_NAME, {"type": "@@quit"})
                    is_waiting_for_response = True
                    continue

                if command == "@@bash":
                    self.client_channel.write_json(
                        CLI_REQUEST_NAME,
                        {
                            "type": "@@bash",
                            "lines": cli_lines
                        }
                    )
                    is_waiting_for_response = True
                    cli_lines = []
                    continue

                if command == "@@python":
                    self.client_channel.write_json(
                        CLI_REQUEST_NAME,
                        {
                            "type": "@@python",
                            "lines": cli_lines
                        }
                    )
                    is_waiting_for_response = True
                    cli_lines = []
                    continue

                if command.startswith("@@mode"):
                    cmds = command.split(" ")
                    if len(cmds) != 2 or cmds[1] not in ("off", "bash", "python"):
                        print("Usage:")
                        print("@@mode off")
                        print("@@mode python")
                        print("@@mode bash")
                    else:
                        line_mode = cmds[1]
                    continue

                if command.startswith("@@log"):
                    cmds = command.split(" ")
                    if len(cmds) != 2:
                        print("Usage:")
                        print("@@log <filename>")
                    else:
                        log_filename = cmds[1]
                    continue

                if command.startswith("@@load"):
                    cmds = command.split(" ")
                    if len(cmds) != 2:
                        print("Usage:")
                        print("@@load <filename>")
                    else:
                        try:
                            with open(cmds[1], "rt") as load_f:
                                for line in load_f:
                                    cli_lines.append(line.rstrip())
                        except Exception as e:
                            print(f"Unable to read from file: {str(e)}")
                    continue

                if command == "@@clear":
                    cli_lines = []
                    continue

                if command == "@@show":
                    for line in cli_lines:
                        print(line)
                    print()
                    continue

                # for any other command, we will append to the cli buffer
                if line_mode == "off":
                    cli_lines.append(command)
                else:
                    self.client_channel.write_json(
                        CLI_REQUEST_NAME,
                        {
                            "type": "@@" + line_mode,
                            "lines": [ command ]
                        }
                    )
                    is_waiting_for_response = True
            else:
                if self.client_channel.has_json(CLI_RESPONSE_NAME):
                    response = self.client_channel.read_json(CLI_RESPONSE_NAME)
                    self.client_channel.delete_json(CLI_RESPONSE_NAME)
                    # print('#################################################')
                    # print('# Response                                      #')
                    # print(f"# status   : {response['status']}")
                    # if 'exit_code' in response:
                    #     print(f"# exit_code: {response['exit_code']}")
                    # print('#################################################')
                    cprint(response['output'], 'green', 'on_red')
                    if log_filename is not None:
                        try:
                            with open(log_filename, "a+t") as log_f:
                                print(response['output'], file=log_f)
                                print("", file=log_f)
                        except Exception as e:
                            print(f"Unable to write to file {log_filename}: {str(e)}")

                    if command == "@@quit":
                        break

                    if self.is_job_active():
                        self.last_job_ok_time = datetime.utcnow()
                        is_waiting_for_response = False
                    else:
                        print("Job quit unexpectedly!")
                        break
                else:
                    time.sleep(1) # do not sleep too long since this is an interactive session
                    now = datetime.utcnow()
                    if self.last_job_ok_time is None or (now - self.last_job_ok_time).total_seconds() >= 10:
                        if self.is_job_active():
                            self.last_job_ok_time = now
                        else:
                            print("Job quit unexpectedly!")
                            break

                    print(f"\r{cli_wait_prompt[cli_wait_prompt_idx]}\r", end="")
                    cli_wait_prompt_idx = (cli_wait_prompt_idx + 1) % 4


def cli_main(spark, args, sysops={}):
    channel = sysops['channel']
    console = PySparkConsole(locals={'spark': spark, 'sysops': sysops})

    channel.write_json(
        spark,
        CLI_RESPONSE_NAME,
        {
            "status": "ok",
            "output": "Welcome to Spark-CLI\n\n" + HELP_TEXT,
        }
    )

    while True:
        if not channel.has_json(spark, CLI_REQUEST_NAME):
            time.sleep(1)
            continue

        user_input = channel.read_json(spark, CLI_REQUEST_NAME)
        channel.delete_json(spark, CLI_REQUEST_NAME)

        if user_input["type"] == "@@quit":
            channel.write_json(
                spark,
                CLI_RESPONSE_NAME,
                {
                    "status": "ok",
                    "output": "Spark-CLI quit gracefully",
                }
            )
            break
        if user_input["type"] == "@@bash":
            handle_bash(spark, user_input, channel)
            continue
        if user_input["type"] == "@@python":
            handle_python(spark, user_input, console, channel)
            continue
    return {"status": "ok"}

SERVER_TO_CLIENT_REQUEST    = "stc-request.json"
SERVER_TO_CLIENT_RESPONSE   = "stc-response.json"

class RemoteCallException(Exception):
    def __init__(self, status, exception_msg=None, exception_name=None):
        if status == 'nohandler':
            msg = 'remote call failed: no handler'
        elif status == 'exception':
            msg = f'remote call failed: exception, name="{exception_name}", msg="{exception_msg}"'
        else:
            msg =f'remote call failed: something is wrong'

        super(RemoteCallException, self).__init__(msg)
        self.msg = msg
        self.status = status
        self.exception_msg = exception_msg
        self.exception_name = exception_name


def server_ask_client(spark, channel, content, timeout=600, check_interval=5):
    """Called by server, ask a question and wait for answer
    """
    channel.write_json(spark, SERVER_TO_CLIENT_REQUEST, content)
    start_time = datetime.utcnow()
    while True:
        if channel.has_json(spark, SERVER_TO_CLIENT_RESPONSE):
            response = channel.read_json(spark, SERVER_TO_CLIENT_RESPONSE)
            channel.delete_json(spark, SERVER_TO_CLIENT_RESPONSE)
            if response['status'] == 'ok':
                return response['answer']
            elif response['status'] == 'nohandler':
                raise RemoteCallException("nohandler")
            elif response['status'] == 'exception':
                exception_name = response.get('exception_name')
                exception_msg = response.get('exception_msg')
                raise RemoteCallException(
                    "exception",
                    exception_name = response.get('exception_name'),
                    exception_msg = response.get('exception_msg')
                )
            else:
                raise RemoteCallException(None)


        if (datetime.utcnow() - start_time).total_seconds() >= timeout:
            raise Exception("Ask is not answered: timed out")
        else:
            time.sleep(check_interval)


def handle_server_ask(channel, handlers):
    """Called by job submitter to answer ask from server with handlers
    """
    if handlers is None:
        return

    if not channel.has_json(SERVER_TO_CLIENT_REQUEST):
        return

    content = channel.read_json(SERVER_TO_CLIENT_REQUEST)
    channel.delete_json(SERVER_TO_CLIENT_REQUEST)
    print(f"handle_server_ask: topic={content.get('topic')}")

    for handler in handlers:
        try:
            handled, out = handler(content)
            if handled:
                channel.write_json(
                    SERVER_TO_CLIENT_RESPONSE,
                    {
                        'status': 'ok',
                        'answer': out
                    }
                )
            return
        except Exception as e:
                traceback.print_exc()
                channel.write_json(
                    SERVER_TO_CLIENT_RESPONSE,
                    {
                        'status': 'exception',
                        "exception_name": e.__class__.__name__,
                        "exception_msg": str(e)
                    }
                )
                return

    channel.write_json(
        SERVER_TO_CLIENT_RESPONSE,
        {
            'status': 'nohandler'
        }
    )
    return




import os
from setuptools import setup, find_packages

# The directory containing this file
HERE = os.path.dirname(os.path.abspath(__file__))

# The text of the README file
with open(os.path.join(HERE, "README.md"), "r") as f:
    README = f.read()

# This call to setup() does all the work
setup(
    name="spark-etl",
    version="0.0.105",
    description="Generic ETL Pipeline Framework for Apache Spark",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/stonezhong/spark_etl",
    author="Stone Zhong",
    author_email="stonezhong@hotmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    package_dir = {'': 'src'},
    packages=find_packages(where='src'),
    install_requires=["requests", "Jinja2", "termcolor"],
    entry_points={
        "console_scripts": [
            "etl=spark_etl.cmds:main",
        ]
    },
)



import importlib
import argparse
import json
import os
import subprocess
import sys
import random
import errno

from pyspark.sql import SparkSession
from pyspark import SparkFiles

random.seed()

class ServerChannel:
    def __init__(self, run_dir):
        self.run_dir = run_dir

    def read_json(self, spark, name):
        with open(os.path.join(self.run_dir, name), "r") as f:
            return json.load(f)

    def has_json(self, spark, name):
        return os.path.isfile(os.path.join(self.run_dir, name))

    def write_json(self, spark, name, payload):
        with open(os.path.join(self.run_dir, name), "w") as f:
            json.dump(payload, f)

    def delete_json(self, spark, name):
        os.remove(os.path.join(self.run_dir, name))

# lib installer
def _install_libs(run_home):
    print(f"job_loader._install_libs: enter, run_home = {run_home}")

    lib_zip     = SparkFiles.get("lib.zip")
    lib_dir     = os.path.join(run_home, "python_libs")
    lock_name   = os.path.join(run_home, '__lock__')

    for i in range(0, 100):
        try:
            lock_fh = os.open(lock_name, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
            os.close(lock_fh)
            try:
                if not os.path.isdir(lib_dir):
                    print("_install_libs: install lib starts")
                    os.makedirs(lib_dir)
                    subprocess.check_call(['unzip', "-qq", lib_zip, "-d", lib_dir])
                    print("_install_libs: install lib done")
                if lib_dir not in sys.path:
                    print(f"_install_libs: add {lib_dir} path")
                    sys.path.insert(0, lib_dir)
                print("job_loader._install_libs: exit")
                return
            finally:
                os.remove(lock_name)
        except OSError as e:
            if e.errno == errno.EEXIST:
                time.sleep(random.randint(1, 10))
                continue
            raise

    raise Exception("Failed to install libraries!")


def _bootstrap():
    parser = argparse.ArgumentParser(description='job')
    parser.add_argument(
        "--run-id", type=str, required=True, help="Run ID",
    )
    parser.add_argument(
        "--run-dir", type=str, required=True, help="Run Directory",
    )
    parser.add_argument(
        "--app-dir", type=str, required=True, help="Application Directory",
    )
    parser.add_argument(
        "--enable-aws-s3",
        help="Allow pyspark to access aws s3 buckets",
        action="store_true",
    )
    parser.add_argument(
        "--aws-account", type=str, required=False, help="AWS Account json file",
    )
    parser.add_argument(
        "--aws-s3-buffer-dir", type=str, required=False, help="AWS S3 buffer dir",
    )
    args = parser.parse_args()
    spark = SparkSession.builder.appName(f"RunJob-{args.run_id}").getOrCreate()

    if args.enable_aws_s3:
        hadoop_conf = spark._jsc.hadoopConfiguration()
        hadoop_conf.set("fs.s3.impl", "org.apache.hadoop.fs.s3native.NativeS3FileSystem")
        if args.aws_s3_buffer_dir is not None:
            hadoop_conf.set("fs.s3.buffer.dir", args.aws_s3_buffer_dir)

        if args.aws_account is not None:
            with open(args.aws_account, "rt") as f:
                aws_account_content = json.load(f)
            hadoop_conf.set("fs.s3.awsAccessKeyId", aws_account_content['aws_access_key_id'])
            hadoop_conf.set("fs.s3.awsSecretAccessKey", aws_account_content['aws_secret_access_key'])


    sc = spark.sparkContext
    sc.addPyFile(os.path.join(args.app_dir, "app.zip"))
    sc.addFile(os.path.join(args.app_dir, 'lib.zip'))

    print(f"run-id:  {args.run_id}")
    print(f"run-dir: {args.run_dir}")
    print(f"app-dir: {args.app_dir}")

    run_home = os.path.join(args.run_dir, args.run_id)
    print(f"run-home: {run_home}")

    # setup lib path
    _install_libs(run_home)

    # load input args
    with open(os.path.join(run_home, "input.json"), "r") as f:
        input_args = json.load(f)


    try:
        entry = importlib.import_module("main")
        result = entry.main(spark, input_args, sysops={
            "install_libs": lambda : _install_libs(run_home),
            "channel": ServerChannel(os.path.join(args.run_dir, args.run_id))
        })

        # save output
        with open(os.path.join(run_home, "result.json"), "w") as out_f:
            json.dump(result, out_f)

    finally:
        spark.stop()

_bootstrap()




class ServerChannelInterface:
    def read_json(self, spark, name):
        raise NotImplementedError()


    def has_json(self, spark, name):
        raise NotImplementedError()


    def write_json(self, spark, name, payload):
        raise NotImplementedError()


    def delete_json(self, spark, name):
        raise NotImplementedError()


class ClientChannelInterface:
    def read_json(self, name):
        raise NotImplementedError()


    def write_json(self, name, payload):
        raise NotImplementedError()


    def has_json(self, name):
        raise NotImplementedError()


    def delete_json(self, name):
        raise NotImplementedError()



import os
import subprocess
import tempfile
from contextlib import contextmanager, suppress
import shutil
import json
from distutils.dir_util import copy_tree

@contextmanager
def _act_in_dir(new_dir):
    current_dir = os.getcwd()
    try:
        os.chdir(new_dir)
        yield
    finally:
        os.chdir(current_dir)

def _delete_file_if_exist(filename):
    if os.path.isfile(filename):
        os.remove(filename)

class Application:
    """Represent a spark application
    """

    def __init__(self, location):
        """
        Parameters
        ----------
        location: str
            The directory path for the source code location of the application. See examples/myapp for example.
        """
        self.location = location
        with open(f"{self.location}/manifest.json", "r") as f:
            self.manifest = json.load(f)

    @property
    def version(self):
        """The application version specified by the manifest file"""
        return self.manifest['version']

    def _create_venv(self):
        tmpdir = tempfile.mkdtemp()
        subprocess.check_call(['/usr/bin/python3', "-m", "venv", tmpdir])
        subprocess.check_call([
            os.path.join(tmpdir, "bin", "python"),
            "-m", "pip",
            "install","pip", "setuptools", "--upgrade"
        ])
        subprocess.check_call([
            os.path.join(tmpdir, "bin", "python"),
            "-m", "pip", "install","wheel"
        ])
        return tmpdir

    def _get_tmp_requirements(self, default_libs):
        with tempfile.NamedTemporaryFile(mode="w+t", delete=False) as f:
            for line in default_libs:
                print(line, file=f)
            app_req_filename = os.path.join(self.location, "requirements.txt")
            if os.path.isfile(app_req_filename):
                with open(app_req_filename, "rt") as arf:
                    for line in arf:
                        print(line, file=f)
            return f.name


    def build(self, destination, default_libs=[]):
        """
        Build the application, it generates the necessary artifacts for the application including:
            app.zip         -- the archive for the application code
            lib.zip         -- the archive for the library
            main.py         -- the main application entry
            manifest.json   -- the application manifest file

        Parameters
        ----------
        location: str
            The directory where those generated artifacts is stored.

        default_libs: list of str
            A list of libraries needed which is not specified in the requirements.txt.
            E.g. ["six==1.11.0"]
        """

        os.makedirs(destination, exist_ok=True)
        build_stage_dir = tempfile.mkdtemp()

        venv_dir = self._create_venv()

        lib_dir = os.path.join(build_stage_dir, "lib")
        os.mkdir(lib_dir)
        tmp_requirements = self._get_tmp_requirements(default_libs)
        subprocess.check_call([
            os.path.join(venv_dir, "bin", "python"),
            "-m", "pip",
            "install", "-r", tmp_requirements,
            "-t", lib_dir
        ])
        os.remove(tmp_requirements)

        # generate archive for lib
        lib_filename = os.path.join(build_stage_dir, 'lib.zip')
        with _act_in_dir(lib_dir):
            subprocess.run(['zip', "-r", lib_filename, "."])

        # generate archive for app
        app_filename = os.path.join(build_stage_dir, 'app.zip')
        with _act_in_dir(self.location):
            subprocess.run(['zip', "-r", app_filename, "."])

        for filename in ["main.py", "manifest.json", "lib.zip", "app.zip"]:
            _delete_file_if_exist(os.path.join(destination, filename))

        shutil.copyfile(lib_filename, os.path.join(destination, 'lib.zip'))
        shutil.copyfile(app_filename, os.path.join(destination, 'app.zip'))
        shutil.copyfile(
            os.path.join(self.location, "manifest.json"),
            os.path.join(destination, 'manifest.json')
        )
        shutil.copyfile(
            os.path.join(self.location, "main.py"),
            os.path.join(destination, 'main.py')
        )
        shutil.rmtree(venv_dir)
        shutil.rmtree(build_stage_dir)



import argparse
import json
import importlib
import os
import glob
from jinja2 import Template

from spark_etl import Application, Build

class Config:
    def __init__(self, config):
        self.config = config

    @classmethod
    def get(cls, config_filename):
        if os.path.isfile(config_filename):
            with open(config_filename, "r") as f:
                config = json.load(f)
        else:
            config = { }
        
        # TODO: validate schema
        return Config(config)
    
    @property
    def apps_dir(self):
        return self.config.get("apps_dir", "apps")
    
    @property
    def builds_dir(self):
        return self.config.get("builds_dir", ".builds")
    
    @property
    def default_libs(self):
        common_req_filename = self.config.get("common_requirements")
        if common_req_filename is None:
            return []
        
        ret = []
        with open(common_req_filename, "rt") as f:
            for line in f:
                if line.startswith("#"):
                    continue
                l = line.strip()
                if len(l) > 0:
                    ret.append(l)
        return ret
    
    @property
    def profiles_dir(self):
        return self.config.get("profiles_dir", ".profiles")


    @property
    def artifacts_dir(self):
        return self.config.get("artifacts_dir", ".artifacts")
    
    def get_app_dir(self, app_name):
        return os.path.join(self.apps_dir, app_name)
    
    def get_build_dir(self, app_name):
        return os.path.join(self.builds_dir, app_name)

    def get_profile(self, profile_name):
        artifacts = load_artifacts(self.artifacts_dir)
        profile_filename = os.path.join(self.profiles_dir, f"{profile_name}.json")

        with open(profile_filename, "rt") as cf:
            profile_template_text = cf.read()
        
        profile_template = Template(profile_template_text)
        profile_text = profile_template.render(artifacts=artifacts)

        profile = json.loads(profile_text)
        return Profile(profile)

class Profile:
    def __init__(self, profile):
        self.profile = profile
        self._deployer = None
        self._job_submitter = None
    
    @property
    def deploy_base(self):
        return self.profile['deploy_base']
    
    def get_deploy_dir(self, app_name):
        return os.path.join(self.deploy_base, app_name)
    
    def get_deploy_version_dir(self, app_name, version):
        return os.path.join(self.deploy_base, app_name, version)

    
    @property
    def job_run_options(self):
        return self.profile.get("job_run_options", {})
    
    @property
    def deployer(self):
        if self._deployer is None:
            deployer_config = self.profile["deployer"]
            class_name  = deployer_config['class']
            module      = importlib.import_module('.'.join(class_name.split(".")[:-1]))
            klass       = getattr(module, class_name.split('.')[-1])
            args        = deployer_config.get("args", [])
            kwargs      = deployer_config.get("kwargs", {})
            self._deployer = klass(*args, **kwargs)
        return self._deployer

    
    @property
    def job_submitter(self):
        if self._job_submitter is None:
            job_submitter_config = self.profile["job_submitter"]
            class_name  = job_submitter_config['class']
            module      = importlib.import_module('.'.join(class_name.split(".")[:-1]))
            klass       = getattr(module, class_name.split('.')[-1])
            args    = job_submitter_config.get("args", [])
            kwargs  = job_submitter_config.get("kwargs", {})
            self._job_submitter = klass(*args, **kwargs)
        return self._job_submitter


def load_artifacts(root_dir):
    ret = {}
    if not os.path.isdir(root_dir):
        return ret
    for filename in os.listdir(root_dir):
        full_filename = os.path.join(root_dir, filename)
        if os.path.isdir(full_filename):
            ret[filename] = load_artifacts(full_filename)
        else:
            with open(full_filename, "rt") as f:
                ret[filename] = json.dumps(f.read())[1:-1]
    return ret


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-a", "--action", required=True, choices=['build', 'deploy', 'run']
    )
    parser.add_argument(
         "-c", "--config-filename", help="Configuration file"
    )
    parser.add_argument(
        "-p", "--app-name", help="Application name"
    )
    parser.add_argument(
        "-f", "--profile-name", help="Profile name"
    )
    parser.add_argument(
        "-v", "--version", help="Application version"
    )
    parser.add_argument(
        "--run-args", help="Arguments for run, a filename to a json"
    )
    parser.add_argument(
        "--cli-mode",
        action="store_true",
        help="Using cli mode?"
    )
    args = parser.parse_args()
    
    config_filename = args.config_filename or "config.json"
    config = Config.get(config_filename)

    if args.action == "build":
        do_build(config, args)
    elif args.action == "deploy":
        do_deploy(config, args)
    elif args.action == "run":
        do_run(config, args)

    return


# build an application
def do_build(config, args):
    if args.app_name is None:
        raise Exception("Please specify app-name!")

    app_name    = args.app_name
    print(f"Building application: {app_name}")
    build_dir   = config.get_build_dir(app_name)
    app_dir     = config.get_app_dir(app_name)
    # cleanup the old build
    os.makedirs(build_dir, exist_ok=True)
    for f in glob.glob(f'{build_dir}/*'):
        os.remove(f)
    app = Application(app_dir)
    app.build(build_dir, default_libs=config.default_libs)
    print("Build application: done!")


def do_deploy(config, args):
    if args.app_name is None:
        raise Exception("Please specify app-name")
    if args.profile_name is None:
        raise Exception("Please specify profile-name")

    app_name        = args.app_name
    profile_name    = args.profile_name
    print(f"Deploy application: {args.app_name}, using profile {profile_name}")

    profile = config.get_profile(profile_name)
    build_dir       = config.get_build_dir(app_name)
    deploy_dir      = profile.get_deploy_dir(app_name)
    profile.deployer.deploy(build_dir, deploy_dir)
    print("Deploy application: done")


def do_run(config, args):
    if args.app_name is None:
        raise Exception("Please specify app-name")
    if args.profile_name is None:
        raise Exception("Please specify profile-name")

    app_name        = args.app_name
    profile_name    = args.profile_name

    build_dir       = config.get_build_dir(app_name)

    if args.version:
        version = args.version
    else:
        with open(os.path.join(build_dir, "manifest.json"), "r") as f:
            version = json.load(f)['version']

    print(f"Run application: {app_name}, version={version}, using profile {profile_name}")

    run_args = args.run_args
    if run_args is None:
        run_args_value = {}
    else:
        with open(run_args, "r") as f:
            run_args_value = json.load(f)

    profile = config.get_profile(profile_name)
    deploy_version_dir      = profile.get_deploy_version_dir(app_name, version)

    ret = profile.job_submitter.run(
        deploy_version_dir,
        options=profile.job_run_options,
        args=run_args_value,
        cli_mode=args.cli_mode
    )
    print("Run application: done!")
    print(f"return = {ret}")
